strong {
  font-weight: 600;
}

code:not(nav.page-toc code):not(:is(h1, h2, h3, h4, h5, h6) code) .pre {
  font-size: 1rem;
}

/* secondary sidebar toc (right side) */
nav.page-toc code {
  border: none;
  background-color: inherit;
}

/* xrefs to types, etc. in breathe doc output */
dl[class] span.std.std-ref {
  font-family: var(--pst-font-family-monospace);
}

/* add visual separator between sections, ignore subsections */
dl[class]:not(.var, .enumerator, .option-list, .field-list, .footnote, .glossary, .simple):not(:last-child) {
  margin-bottom: 1rem;
  padding-bottom: 2rem;
  border-bottom: 1px solid var(--pst-color-border);
}

dl[class]:not(.var, .enumerator, .option-list, .field-list, .footnote, .glossary, .simple):not(:last-child) > dd {
  margin-left: 1rem;
  padding-top: 1rem;
}
