/*
 * Copyright (c) Advanced Micro Devices, Inc. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "amd_smi/impl/scoped_fd.h"

#include <fcntl.h>

#include <cstring>
#include <string>

#include "rocm_smi/rocm_smi_logger.h"

ScopedFD::ScopedFD(const std::string& path, int flags)
    : fd_(open(path.c_str(), flags)), path_(path) {
  std::ostringstream ss;
  if (fd_ < 0) {
    ss << __PRETTY_FUNCTION__ << " | Failed to open file: " << path_
       << " | Error: " << strerror(errno);
    LOG_ERROR(ss);
  }
}

ScopedFD::~ScopedFD() {
  if (fd_ >= 0) {
    close(fd_);
  }
}

ScopedFD::ScopedFD(ScopedFD&& other) noexcept : fd_(other.fd_), path_(std::move(other.path_)) {
  other.fd_ = -1;
}

ScopedFD& ScopedFD::operator=(ScopedFD&& other) noexcept {
  if (this != &other) {
    if (fd_ >= 0) close(fd_);
    fd_ = other.fd_;
    path_ = std::move(other.path_);
    other.fd_ = -1;
  }
  return *this;
}

int ScopedFD::get() const { return fd_; }

bool ScopedFD::valid() const { return fd_ >= 0; }

ScopedFD::operator int() const { return fd_; }
